package interpreter

import scala.language.implicitConversions

implicit def toResult[V](v: V): Result[V] = Result.Value(v)
enum Result[+V]:
  case Value(v: V)
  case InterpreterError(e: ErrorType) extends Result[Nothing]
  def success = this match
    case Value(v)            => true
    case InterpreterError(e) => false

enum ErrorType:
  case DivisionByZero
  case UnboundVariable(name: String)
  case NumOpOnNonNum
  case PairOpOnNonPair
  case IfCondNotNum

type StrictResult = Result[StrictValue]
enum StrictValue:
  case NumV(n: Int)
  case PairV(first: StrictValue, second: StrictValue)

trait Interpreter:
  type EvalResult
  type Env

  def evaluate(e: Expr)(using Env): EvalResult
  def topLevelEvaluate(e: Expr): StrictResult
