package interpreter
import scala.language.adhocExtensions

import Result.*
import Expr.*
import NumOp.*
import PairOp.*

class LazyInterpreter extends Interpreter:
  enum LazyValue:
    case NumV(n: Int)
    case PairV(first: DeferredExpr, second: DeferredExpr)

  case class DeferredExpr(e: Expr)(using Env):
    lazy val result: LazyResult = evaluate(e)

  type LazyResult = Result[LazyValue]
  type EvalResult = LazyResult
  type Env = Map[String, DeferredExpr]

  import LazyValue.*

  def evaluate(e: Expr)(using env: Env): LazyResult = e match
    case Num(n)              => NumV(n)
    case NumExpr(op, e1, e2) =>
      ???
    case Pair(e1, e2) =>
      PairV(DeferredExpr(e1), DeferredExpr(e2))
    case PairExpr(op, e) =>
      ???
    case IsPair(e) =>
      ???
    case Var(name) =>
      env.get(name) match
        case Some(v) => v.result
        case None    => InterpreterError(ErrorType.UnboundVariable(name))
    case Let(name, e, body) =>
      ???
    case If(cond, thenb, elseb) =>
      ???

  def force(e: LazyResult): StrictResult = e match
    case Value(NumV(n)) => StrictValue.NumV(n)
    case Value(PairV(v1, v2)) =>
      (force(v1.result), force(v2.result)) match
        case (e: InterpreterError, _) => e
        case (_, e: InterpreterError) => e
        case (Value(v1), Value(v2))   => StrictValue.PairV(v1, v2)
    case InterpreterError(e) => InterpreterError(e)

  def topLevelEvaluate(e: Expr): StrictResult =
    force(evaluate(e)(using Map.empty))
