package interpreter

import Result.*
import StrictValue.*
import Expr.*
import NumOp.*
import PairOp.*

class StrictInterpreter extends Interpreter:
  type EvalResult = StrictResult
  type Env = Map[String, StrictValue]

  def evaluate(e: Expr)(using env: Env): StrictResult = e match
    case Num(n) => NumV(n)
    case NumExpr(op, e1, e2) =>
      (evaluate(e1), evaluate(e2)) match
        case (e: InterpreterError, _) => e
        case (_, e: InterpreterError) => e
        case (Value(NumV(n1)), Value(NumV(n2))) =>
          op match
            case Add => NumV(n1 + n2)
            case Sub => NumV(n1 - n2)
            case Mul => NumV(n1 * n2)
            case Div =>
              if n2 == 0 then InterpreterError(ErrorType.DivisionByZero)
              else NumV(n1 / n2)
        case _ => InterpreterError(ErrorType.NumOpOnNonNum)
    case Pair(e1, e2) =>
      (evaluate(e1), evaluate(e2)) match
        case (e: InterpreterError, _) => e
        case (_, e: InterpreterError) => e
        case (Value(v1), Value(v2))   => PairV(v1, v2)
    case PairExpr(op, e) =>
      ???
    case IsPair(e) =>
      ???
    case Var(name) =>
      env.get(name) match
        case Some(v) => v
        case None    => InterpreterError(ErrorType.UnboundVariable(name))
    case Let(name, e, body) =>
      evaluate(e) match
        case e: InterpreterError => e
        case Value(v)            => evaluate(body)(using env + (name -> v))
    case If(cond, thenb, elseb) =>
      ???

  def topLevelEvaluate(e: Expr): StrictResult = evaluate(e)(using Map.empty)
