package interpreter

import scala.language.adhocExtensions
import org.scalacheck.*
import Gen.*
import Prop.*
import interpreter.Result.*
import interpreter.StrictValue.*
import ExprGen.*
import Expr.*

class TestInterpreterProperties extends Properties("Property-based tests"):
  given Arbitrary[Expr] = Arbitrary(
    sized { size =>
      for
        tp <- randomType(size)
        e <- genExpr(size, tp)(using Map.empty)
      yield e
    }
  )

  propertyWithSeed(
    "The strict interpreter does not produce errors on safe programs. (6pts)",
    Some("6H2Co4OvrQM6P_rfxPbMP3ePtG7q4XzD13Rcg2JcC-J=")
  ) =
    val strictEval = StrictInterpreterWithCounter()
    forAll { (e: Expr) =>
      strictEval.topLevelEvaluate(e).success
    }

  propertyWithSeed(
    "The lazy interpreter agrees with the strict interpreter on safe programs and recurses fewer times. (14pts)",
    Some("t4PbCViPj_7eTHK5ugYpZ93GwuUNrN-IT1R1dTFpM0F=")
  ) =
    forAll { (e: Expr) =>
      val strictEval = StrictInterpreterWithCounter()
      val lazyEval = LazyInterpreterWithCounter()

      val e1 = Let("_0", Pair(Num(1), e), PairExpr(PairOp.Second, Var("_0")))

      val r1 = strictEval.topLevelEvaluate(e1)
      val r2 = lazyEval.topLevelEvaluate(e1)

      r1.success && (r1 == r2) && (lazyEval.counter > 0) && (lazyEval.counter < strictEval.counter)
    }
